// license:BSD-3-Clause
// copyright-holders:Olivier Galibert

#include "thom_dsk.h"
#include "ioprocs.h"
#include "osdcore.h"

#define VERBOSE 1

#ifndef LOG_FORMATS
#define LOG_FORMATS(...) do { if (VERBOSE) osd_printf_info(__VA_ARGS__); } while (false)
#endif

thomson_format::thomson_format() : wd177x_format(formats)
{
}

const char *thomson_format::name() const noexcept
{
	return "thomson_fd";
}

const char *thomson_format::description() const noexcept
{
	return "Thomson FD disk image";
}

const char *thomson_format::extensions() const noexcept
{
	return "fd";
}

// 1280K .fd images exist but are not supported. They represent a notional type
// of 4-sided disk that can be inserted into 2 drives at once.
const thomson_format::format thomson_format::formats[] = {
	//
	// 5.25" form factor
	//
	{
		floppy_image::FF_525, floppy_image::DSDD, floppy_image::MFM,
		2000,
		16, 40, 2,
		256, {},
		1, {},
		31, 22, 44
	},
	{	// DS, DD on side 0, SD on side 1
		floppy_image::FF_525, floppy_image::DSDD, floppy_image::MFM,
		2000,
		16, 40, 2,
		256, {},
		1, {},
		31, 22, 44
	},
	{
		floppy_image::FF_525, floppy_image::SSDD, floppy_image::MFM,
		2000,
		16, 40, 1,
		256, {},
		1, {},
		31, 22, 44
	},
	{
		floppy_image::FF_525, floppy_image::DSSD, floppy_image::FM,
		4000,
		16, 40, 2,
		128, {},
		1, {},
		27, 11, 27
	},
	{	// DS, SD on side 0, DD on side 1
		floppy_image::FF_525, floppy_image::DSSD, floppy_image::FM,
		4000,
		16, 40, 2,
		128, {},
		1, {},
		27, 11, 27
	},
	{
		floppy_image::FF_525, floppy_image::SSSD, floppy_image::FM,
		4000,
		16, 40, 1,
		128, {},
		1, {},
		27, 11, 27
	},
	//
	// 3.5" form factor
	//
	{
		floppy_image::FF_35, floppy_image::DSDD, floppy_image::MFM,
		2000,
		16, 80, 2,
		256, {},
		1, {},
		31, 22, 44
	},
	{	// DS, DD on side 0, SD on side 1
		floppy_image::FF_35, floppy_image::DSDD, floppy_image::MFM,
		2000,
		16, 80, 2,
		256, {},
		1, {},
		31, 22, 44
	},
	{
		floppy_image::FF_35, floppy_image::SSDD, floppy_image::MFM,
		2000,
		16, 80, 1,
		256, {},
		1, {},
		31, 22, 44
	},
	{	// DS, SD on side 0, DD on side 1
		floppy_image::FF_35, floppy_image::DSSD, floppy_image::FM,
		4000,
		16, 80, 2,
		128, {},
		1, {},
		17, 12, 22
	},
	{
		floppy_image::FF_35, floppy_image::DSSD, floppy_image::FM,
		4000,
		16, 80, 2,
		128, {},
		1, {},
		17, 12, 22
	},
	{
		floppy_image::FF_35, floppy_image::SSSD, floppy_image::FM,
		4000,
		16, 80, 1,
		128, {},
		1, {},
		17, 12, 22
	},
	{}
};

bool thomson_format::validate_fat(util::random_read &io, const format &f, int offset) const
{
	// Basic Microsoft FAT representation (track 20, sector 2)
	//   $ff: unallocated
	//   $fe: reserved
	//   <$c0: allocated block
	//   >$c0 and <$c9: last block
	int nblocks = (f.track_count * 2 < f.sector_base_size - 1 ? f.track_count * 2 : f.sector_base_size - 1);
	uint8_t fat[256];
	read_at(io, offset + (20 * f.sector_count + 1) * f.sector_base_size, fat, f.sector_base_size);
	// offset 0 is always zero
	if (fat[0])
		return false;
	 // offsets 41 and 42 are reserved: two blocks for FAT
	if (fat[41] != 0xfe)
		return false;
	if (fat[42] != 0xfe)
		return false;
	for(int i = 1; i <= nblocks; i++) {
		if (fat[i] > 0xc8 && fat[i] < 0xfe)
			return false;
		// - value $c0 is not possible (the last block contains at least one sector)
		// - value should be bound by the number of blocks
		if (fat[i] >= nblocks && fat[i] < 0xc1)
			return false;
	}
	return true;
}

bool thomson_format::validate_catalog(util::random_read &io, const format &f, int offset) const
{
	// Basic Microsoft catalog representation (track 20, sector 3 to 15)
	// A file descriptor uses 32 bytes:
	//  - offset 0: (0 unallocated entry, $20-$7f allocated entry, $ff end of catalog
	//  - offset 0-7: filename, left centered, completed by spaces
	//  - offset 8-10: extension
	//  - offset 11: file type (0, 1, 2, 3)
	//  - offset 12: $ff for ascii, 0 for binary
	//  - offset 13: first block
	//  - offset 14-15: number of bytes in the last sector
	//  - offset 16-23: comments
	//  - offset 24-31: reserved

	int nblocks = (f.track_count * 2 < f.sector_base_size - 1 ? f.track_count * 2 : f.sector_base_size - 1);
	bool end_of_catalog = false;
	for(int sect=2; sect < 16; sect++) {
		uint8_t cat[256];
		read_at(io, offset + (20 * f.sector_count + sect) * f.sector_base_size, cat, f.sector_base_size);
		for(int i=0; i < f.sector_base_size; i += 32) {
			bool end_of_filename = false;
			bool end_of_suffix = false;
			if (end_of_catalog) {
				if (cat[i] < 0xff)
					return false;
				continue;
			}
			if (cat[i] == 0)
				continue;
			if (cat[i] == 0xff) {
				end_of_catalog = true;
				continue;
			}
			for(int j=i; j < i+8; j++) {
				if (cat[j] < 0x20 || cat[j] > 0x7f)
					return false;
				if (end_of_filename && cat[j] != 0x20)
					return false;
				end_of_filename = (cat[j] == 0x20);
			}
			for(int j=i+8; j < i+11; j++) {
				if (cat[j] < 0x20 || cat[j] > 0x7f)
					return false;
				if (end_of_suffix && cat[j] != 0x20)
					return false;
				end_of_suffix = (cat[j] == 0x20);
			}
			if (cat[i + 11] > 3)
				return false;
			if (cat[i + 12] > 0 && cat[i + 12] < 0xff)
				return false;
			if (cat[i + 13] >= nblocks)
				return false;
		}
	}
	return true;
}

int thomson_format::find_struct(util::random_read &io, uint32_t form_factor, const std::vector<uint32_t> &variants) const
{
	uint64_t size;
	if(io.length(size))
		return -1;

	for(int i=0; formats[i].form_factor; i++) {
		const format &f = get_track_format(formats[i], 0, 0);
		int track_size = 0;
		int head;
		if(form_factor != floppy_image::FF_UNKNOWN && form_factor != f.form_factor)
			continue;
		if(!variants.empty() && !has_variant(variants, f.variant))
			continue;
		for(head=0; head < f.head_count; head++) {
			const format &tf = get_track_format(formats[i], head, 0);
			if(!validate_fat(io, tf, track_size))
				break;
			if(!validate_catalog(io, tf, track_size))
				break;
			track_size += (tf.track_count * tf.sector_base_size * tf.sector_count);
		}
		if(head < f.head_count)
			continue;
		if(size != track_size)
			continue;
		for(head=0; head < f.head_count; head++) {
			const format &tf = get_track_format(formats[i], head, 0);
			LOG_FORMATS("thomson_format::find_struct: identified %s, %s, %s for head %d with size %d\n",
					floppy_image::get_form_factor_name(tf.form_factor),
					tf.variant == floppy_image::SSDD || tf.variant == floppy_image::DSDD ? "DD" : "SD",
					floppy_image::get_encoding_name(tf.encoding), head,
					tf.track_count * tf.sector_base_size * tf.sector_count);
		}
		return i;
	}
	LOG_FORMATS("thomson_format::find_struct: unidentified for size %d\n", size);
	return -1;
}

int thomson_format::find_size(util::random_read &io, uint32_t form_factor, const std::vector<uint32_t> &variants) const
{
	uint64_t size;
	if(io.length(size))
		return -1;

	for(int i=0; formats[i].form_factor; i++) {
		const format &f = get_track_format(formats[i], 0, 0);
		int track_size = 0;
		if(form_factor != floppy_image::FF_UNKNOWN && form_factor != f.form_factor)
			continue;
		for(int head=0; head < f.head_count; head++) {
			const format &tf = get_track_format(formats[i], head, 0);
			track_size += (tf.track_count * tf.sector_base_size * tf.sector_count);
		}
		if(size != track_size)
			continue;
		for(int head=0; head < f.head_count; head++) {
			const format &tf = get_track_format(formats[i], head, 0);
			LOG_FORMATS("thomson_format::find_size: identified %s, %s, %s for head %d with size %d\n",
					floppy_image::get_form_factor_name(tf.form_factor),
					tf.variant == floppy_image::SSDD || tf.variant == floppy_image::DSDD ? "DD" : "SD",
					floppy_image::get_encoding_name(tf.encoding), head,
					tf.track_count * tf.sector_base_size * tf.sector_count);
		}
		return i;
	}
	LOG_FORMATS("thomson_format::find_size: unidentified for size %d\n", size);
	return -1;
}

int thomson_format::identify(util::random_read &io, uint32_t form_factor, const std::vector<uint32_t> &variants) const
{
	int ret = 0;

	if(find_size(io, form_factor, variants) != -1)
		ret |= FIFID_SIZE;
	if(find_struct(io, form_factor, variants) != -1)
		ret |= FIFID_STRUCT;

	return ret;
}

int thomson_format::get_image_offset(const format &f, int head, int track) const
{
	int offset = 0;

	for(int hd=0; hd < head; hd++) {
		for(int trk=0; trk < f.track_count; trk++) {
			const format &tf = get_track_format(f, hd, trk);
			offset += compute_track_size(tf);
		}
	}

	for(int trk=0; trk < track; trk++) {
		const format &tf = get_track_format(f, head, trk);
		offset += compute_track_size(tf);
	}

	return offset;
}

const wd177x_format::format &thomson_format::get_track_format(const format &f, int head, int track) const
{
	int n = -1;
	for(int i = 0; formats[i].form_factor; i++) {
		if(&formats[i] == &f) {
			n = i;
			break;
		}
	}
	if(n < 0) {
		LOG_FORMATS("thomson_format::get_track_format: Error format not found\n");
		return f;
	}
	if(head >= f.head_count) {
		LOG_FORMATS("thomson_format::get_track_format: Error invalid head %d\n", head);
		return f;
	}
	if(track >= f.track_count) {
		LOG_FORMATS("thomson_format::get_track_format: Error invalid track %d\n", track);
		return f;
	}
	for(int i = 0; i < 12; i += 6) {	// both form factors
		if(n == i + 1 && head == 1)	// DS, DD on side 0, SD on side 1
			n = i + 3;
		if(n == i + 4 && head == 1)	// DS, SD on side 0, DD on side 1
			n = i + 0;
	}
	return formats[n];
}

floppy_image_format_t::desc_e* thomson_format::get_desc_fm(const format &f, int &current_size, int &end_gap_index) const
{
	// interleave value (set in DK.NUM at $604D for DKFORM operation):
	//  - 4 for CD90-015 TO7 and MO5 Disk Basic 1.0
	//  - 5 for CD90-640 TO7 Disk Basic 1.2
	//  - 7 for CD90-640 MO5 Disk Basic 1.1
	//  - 5 for CD90-351 TO7 3"1/2 Disk Basic 1.1
	//  - 7 for CD90-351 MO5 3"1/2 Disk Basic 1.0
	//  - 7 for Basic 128 and Basic 512
	static floppy_image_format_t::desc_e desc[24] = {
		/* 00 */ { SECTOR_INTERLEAVE_SKEW, 4, 0}, // we set an interleave value
		/* 01 */ { FM, 0xff, 0 },
		/* 02 */ { SECTOR_LOOP_START, 0, 0 },
		/* 03 */ {   FM, 0x00, 6 },
		/* 04 */ {   CRC_CCITT_FM_START, 1 },
		/* 05 */ {     RAW, 0xf57e, 1 },
		/* 06 */ {     TRACK_ID_FM },
		/* 07 */ {     FM, 0x00, 1 },	// the head id is always zero
		/* 08 */ {     SECTOR_ID_FM },
		/* 09 */ {     SIZE_ID_FM },
		/* 10 */ {   CRC_END, 1 },
		/* 11 */ {   CRC, 1 },
		/* 12 */ {   FM, 0xff, 0 },
		/* 13 */ {   FM, 0x00, 6 },
		/* 14 */ {   CRC_CCITT_FM_START, 2 },
		/* 15 */ {     RAW, 0xf56f, 1 },
		/* 16 */ {     SECTOR_DATA_FM, -1 },
		/* 17 */ {   CRC_END, 2 },
		/* 18 */ {   CRC, 2 },
		/* 19 */ {   FM, 0xff, 0 },
		/* 20 */ { SECTOR_LOOP_END },
		/* 21 */ { FM, 0xff, 0 },
		/* 22 */ { RAWBITS, 0xffff, 0 },
		/* 23 */ { END }
	};

	desc[1].p2 = f.gap_1;
	desc[2].p2 = f.sector_count - 1;
	desc[12].p2 = f.gap_2;
	desc[19].p2 = f.gap_3;

	current_size = f.gap_1*16;
	if(f.sector_base_size)
		current_size += f.sector_base_size * f.sector_count * 16;
	else {
		for(int j=0; j != f.sector_count; j++)
			current_size += f.per_sector_size[j] * 16;
	}
	current_size += (6+1+4+2+f.gap_2+6+1+2+f.gap_3) * f.sector_count * 16;

	end_gap_index = 21;

	return desc;
}

floppy_image_format_t::desc_e* thomson_format::get_desc_mfm(const format &f, int &current_size, int &end_gap_index) const
{
	static floppy_image_format_t::desc_e desc[26] = {
		/* 00 */ { SECTOR_INTERLEAVE_SKEW, 6, 0}, // we set an interleave value
		/* 01 */ { MFM, 0x4e, 0 },
		/* 02 */ { SECTOR_LOOP_START, 0, 0 },
		/* 03 */ {   MFM, 0x00, 12 },
		/* 04 */ {   CRC_CCITT_START, 1 },
		/* 05 */ {     RAW, 0x4489, 3 },
		/* 06 */ {     MFM, 0xfe, 1 },
		/* 07 */ {     TRACK_ID },
		/* 08 */ {     FM, 0x00, 1 }, // the head id is always zero
		/* 09 */ {     SECTOR_ID },
		/* 10 */ {     SIZE_ID },
		/* 11 */ {   CRC_END, 1 },
		/* 12 */ {   CRC, 1 },
		/* 13 */ {   MFM, 0x4e, 0 },
		/* 14 */ {   MFM, 0x00, 12 },
		/* 15 */ {   CRC_CCITT_START, 2 },
		/* 16 */ {     RAW, 0x4489, 3 },
		/* 17 */ {     MFM, 0xfb, 1 },
		/* 18 */ {     SECTOR_DATA, -1 },
		/* 19 */ {   CRC_END, 2 },
		/* 20 */ {   CRC, 2 },
		/* 21 */ {   MFM, 0x4e, 0 },
		/* 22 */ { SECTOR_LOOP_END },
		/* 23 */ { MFM, 0x4e, 0 },
		/* 24 */ { RAWBITS, 0x9254, 0 },
		/* 25 */ { END }
	};

	desc[1].p2 = f.gap_1;
	desc[2].p2 = f.sector_count - 1;
	desc[13].p2 = f.gap_2;
	desc[21].p2 = f.gap_3;

	current_size = f.gap_1*16;
	if(f.sector_base_size)
		current_size += f.sector_base_size * f.sector_count * 16;
	else {
		for(int j=0; j != f.sector_count; j++)
			current_size += f.per_sector_size[j] * 16;
	}
	current_size += (12+3+1+4+2+f.gap_2+12+3+1+2+f.gap_3) * f.sector_count * 16;

	end_gap_index = 23;

	return desc;
}

bool thomson_format::load(util::random_read &io, uint32_t form_factor, const std::vector<uint32_t> &variants, floppy_image &image) const
{
	int const type = find_size(io, form_factor, variants);
	if(type == -1)
		return false;

	const format &f = formats[type];
	int max_tracks, max_heads;
	image.get_maximal_geometry(max_tracks, max_heads);

	if(f.track_count > max_tracks) {
		osd_printf_error("thomson_format: Number of tracks in image file too high for floppy drive (%d > %d)\n", f.track_count, max_tracks);
		return false;
	}

	for(int track=0; track < f.track_count; track++)
		for(int head=0; head < f.head_count; head++) {
			uint8_t sectdata[40*512];
			desc_s sectors[40];
			floppy_image_format_t::desc_e *desc;
			int current_size;
			int end_gap_index;
			const format &tf = get_track_format(f, head, track);

			switch (tf.encoding)
			{
			case floppy_image::FM:
				desc = get_desc_fm(tf, current_size, end_gap_index);
				break;
			case floppy_image::MFM:
			default:
				desc = get_desc_mfm(tf, current_size, end_gap_index);
				break;
			}

			int total_size = 200000000/tf.cell_size;
			int remaining_size = total_size - current_size;
			if(remaining_size < 0) {
				osd_printf_error("thomson_format: Incorrect track layout, max_size=%d, current_size=%d\n", total_size, current_size);
				return false;
			}

			// Fixup the end gap
			desc[end_gap_index].p2 = remaining_size / 16;
			desc[end_gap_index + 1].p2 = remaining_size & 15;
			desc[end_gap_index + 1].p1 >>= 16-(remaining_size & 15);

			build_sector_description(tf, sectdata, sectors, track, head);
			int track_size = compute_track_size(tf);
			/*auto const [err, actual] =*/ read_at(io, get_image_offset(f, head, track), sectdata, track_size); // FIXME: check for errors and premature EOF
			generate_track(desc, track, head, sectors, tf.sector_count, total_size, image);
		}

	image.set_form_variant(f.form_factor, f.variant);

	return true;
}

const thomson_format FLOPPY_THOMSON_FORMAT;
